/* $Id: qodem.c,v 1.180 2010/10/06 12:19:51 lamonte Exp $
 *
 * qodem.c
 *
 * Additional acknowledgements, news, and updates to the Qodem Project
 * are available at http://qodem.sourceforge.net .
 *
 * This module is licensed under the GNU General Public License Version 2.
 * Please see the file "COPYING" in this directory for more information
 * about the GNU General Public License Version 2.
 *
 *     Copyright (C) 2003  Kevin Lamonte
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */

#include "config.h"
#include "common.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>	/* vsprintf */
#include <getopt.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <pwd.h>
#include "qodem.h"
#include "input.h"
#include "colors.h"
#include "screen.h"
#include "protocols.h"
#include "console.h"
#include "keyboard.h"
#include "translate.h"
#include "modem.h"
#include "music.h"
#include "forms.h"
#include "linux.h"
#include "states.h"
#include "status.h"
#include "options.h"
#include "emulation.h"
#include "phonebook.h"
#include "dialer.h"
#include "script.h"

/* #define DEBUG_IO 1 */
#undef DEBUG_IO
#ifdef DEBUG_IO
static FILE * DEBUG_IO_HANDLE = NULL;
#endif

/*
 * Define this to enable random line noise on both sides of the link.
 * There is a random chance that 1 in every line_noise_per_bytes is
 * turned into crap.
 */
/* #define LINE_NOISE */
#undef LINE_NOISE
#ifdef LINE_NOISE
static const int line_noise_per_bytes = 10000;
static Q_BOOL noise_stop = Q_FALSE;
#endif /* LINE_NOISE */

/* Global exit return code */
static int q_exitrc = EXIT_OK;

/* Global child TTY name */
char * q_child_ttyname = NULL;

/* Global child tty file descriptor */
int q_child_tty_fd = -1;

/* Global child process ID */
pid_t q_child_pid = -1;

/* Global status struct */
struct q_status_struct q_status;

/* Global width value */
int WIDTH;

/* Global height value */
int HEIGHT;

/* Global status height value */
int STATUS_HEIGHT;

/* Global base directory */
char * q_home_directory = NULL;

/* Global screensaver timeout */
int q_screensaver_timeout;

/* Global keepalive timeout */
int q_keepalive_timeout;
char q_keepalive_bytes[128];
unsigned int q_keepalive_bytes_n;

/* Raw input buffer */
static unsigned char q_buffer_raw[Q_BUFFER_SIZE];
static int q_buffer_raw_n = 0;

/* Transfer output buffer */
static unsigned char q_transfer_buffer_raw[Q_BUFFER_SIZE];
static int q_transfer_buffer_raw_n = 0;

/* These are used by the select() call in data_handler() */
static fd_set readfds;			/* read FD_SET */
static fd_set writefds;			/* write FD_SET */
static fd_set exceptfds;		/* exception FD_SET */

/* The last time we saw data. */
static time_t data_time;

/* The last time we sent data. */
time_t q_data_sent_time;

/* The initial call to make as requested by the command line arguments */
static struct q_phone_struct initial_call;

/* Command-line options */
static struct option q_getopt_long_options[] = {
	{"connect",		1,	0,	0},
	{"connect-method",	1,	0,	0},
	{"enable-capture",	1,	0,	0},
	{"enable-logging",	1,	0,	0},
	{"help",		0,	0,	0},
	{"username",		1,	0,	0},
	{"version",		0,	0,	0},
	{0,			0,	0,	0}
};

/* All outgoing data goes through here */
int qodem_write(const int fd, char * data, const int data_n, Q_BOOL sync) {
	int i;
	int rc;
	int old_errno;

	/* Mark/space parity */
	if ((q_status.serial_open == Q_TRUE) && (q_serial_port.parity == PARITY_MARK)) {
		/* Outgoing data as MARK parity:  set the 8th bit */
		for (i = 0; i < data_n; i++) {
			data[i] |= 0x80;
		}
	}
	if ((q_status.serial_open == Q_TRUE) && (q_serial_port.parity == PARITY_SPACE)) {
		/* Outgoing data as SPACE parity:  strip the 8th bit */
		for (i = 0; i < data_n; i++) {
			data[i] &= 0x7F;
		}
	}

#ifdef DEBUG_IO
	fprintf(DEBUG_IO_HANDLE, "qodem_write() OUTPUT bytes: ");
	for (i = 0; i < data_n; i++) {
		fprintf(DEBUG_IO_HANDLE, "%02x ", data[i] & 0xFF);
	}
	fprintf(DEBUG_IO_HANDLE, "\n");
	fflush(DEBUG_IO_HANDLE);
#endif

	/* Write bytes out */
	rc = write(fd, data, data_n);
	old_errno = errno;
	if (rc < 0) {
#ifdef DEBUG_IO
		fprintf(DEBUG_IO_HANDLE, "qodem_write() write() error %s (%d)\n", strerror(errno), errno);
		fflush(DEBUG_IO_HANDLE);
#endif
	}

	if (sync == Q_TRUE) {
		/* Sync the stream */
		if (fsync(fd) != 0) {
#ifdef DEBUG_IO
			fprintf(DEBUG_IO_HANDLE, "qodem_write() fsync() error %s (%d)\n", strerror(errno), errno);
			fflush(DEBUG_IO_HANDLE);
#endif
		}
	}

	/* Reset clock for keepalive/idle timeouts */
	time(&q_data_sent_time);

	errno = old_errno;
	return rc;
} /* ---------------------------------------------------------------------- */

static char * usage_string() {
	return _(""
"'qodem' is a terminal connection manager with scrollback, capture,\n"
"and basic scripting support.\n"
"\n"
"Usage: qodem [OPTIONS] { [--connect] | [command line] }\n"
"\n"
"Options:\n"
"\n"
"If a long option shows an argument as mandatory, then it is mandatory\n"
"for the equivalent short option also.  Similarly for optional arguments.\n"
"\n"
"      --connect HOST                  Immediately open a connection to "
"HOST.\n"
"                                      The default connection method is \"ssh"
"\".\n"
"      --connect-method METHOD         Use METHOD to connect for the --"
"connect\n"
"                                      option.  Valid values are\n"
"                                      \"ssh\", \"rlogin\", \"rsh\", \"telnet,"
"\",\n"
"                                      and \"shell\".\n"
"      --enable-capture FILENAME       Capture the entire session and save "
"to\n"
"                                      FILENAME.\n"
"      --enable-logging FILENAME       Enable the session log and save to\n"
"                                      FILENAME.\n"
"  -h, --help                          This help screen\n"
"      --username USERNAME             Log in as USERNAME\n"
"      --version                       Version\n"
"\n"
"qodem can also open a raw shell with the command line given.\n"
"So for example 'qodem --connect my.host --connect-method ssh' is\n"
"equivalent to 'qodem ssh my.host' .  The --connect option cannot\n"
"be used if a command line is specified.\n"
"\n");
} /* ---------------------------------------------------------------------- */

/*
 * check_for_help - see if the user asked for help or version information
 */
static int check_for_help(int argc, char * const argv[]) {
	int i;

	for (i = 0; i < argc; i++) {

		/* Special case: help means exit */
		if (strncmp(argv[i], "--help", strlen("--help")) == 0) {
			printf("%s", usage_string());
			return EXIT_HELP;
		}
		if (strncmp(argv[i], "-h", strlen("-h")) == 0) {
			printf("%s", usage_string());
			return EXIT_HELP;
		}
		if (strncmp(argv[i], "-?", strlen("-?")) == 0) {
			printf("%s", usage_string());
			return EXIT_VERSION;
		}

	}

	/* The user did not ask for help */
	return 0;
} /* ---------------------------------------------------------------------- */

/*
 * process_command_line_option
 */
static void process_command_line_option(const char * option, const char * value) {

	wchar_t value_wchar[128];

	/*
	 fprintf(stdout, "OPTION=%s VALUE=%s\n", option, value);
	 */

	/* Special case: help means exit */
	if (strncmp(option, "help", strlen("help")) == 0) {
		printf("%s", usage_string());
		q_program_state = Q_STATE_EXIT;
	}

	/* Special case: version means exit */
	if (strncmp(option, "version", strlen("version")) == 0) {
		printf("%s", _(""
"qodem version 0.3.2\n"
"Copyright(c) 2010 Kevin Lamonte\n"
"\n"
"This program is free software; you can redistribute it and/or modify\n"
"it under the terms of the GNU General Public License as published by\n"
"the Free Software Foundation; either version 2 of the License, or\n"
"(at your option) any later version.\n"
"\n"));
		q_program_state = Q_STATE_EXIT;
	}

	if (strncmp(option, "enable-capture", strlen("enable-capture")) == 0) {
		start_capture(value);
	}

	if (strncmp(option, "enable-logging", strlen("enable-logging")) == 0) {
		start_logging(value);
	}

	if (strcmp(option, "connect") == 0) {
		initial_call.address = (char *)value;
	}

	if (strncmp(option, "username", strlen("username")) == 0) {
		memset(value_wchar, 0, sizeof(value_wchar));
		mbstowcs(value_wchar, value, strlen(value));
		initial_call.username = Xwcsdup(value_wchar, __FILE__, __LINE__);
	}

	if (strncmp(option, "connect-method", strlen("connect-method")) == 0) {
		if (strncmp(value, "ssh", strlen("ssh")) == 0) {
			initial_call.method = Q_DIAL_METHOD_SSH;
		} else if (strncmp(value, "shell", strlen("shell")) == 0) {
			initial_call.method = Q_DIAL_METHOD_SHELL;
			initial_call.address = "";
		} else if (strncmp(value, "rlogin", strlen("rlogin")) == 0) {
			initial_call.method = Q_DIAL_METHOD_RLOGIN;
		} else if (strncmp(value, "rsh", strlen("rsh")) == 0) {
			initial_call.method = Q_DIAL_METHOD_RSH;
		} else if (strncmp(value, "telnet", strlen("telnet")) == 0) {
			initial_call.method = Q_DIAL_METHOD_TELNET;
		/*
		} else if (strncmp(value, "tn3270", strlen("tn3270")) == 0) {
			initial_call.method = Q_DIAL_METHOD_TN3270;
		 */
		}
	}


} /* ---------------------------------------------------------------------- */

/*
 * qlog - emit a line to the session log
 */
void qlog(const char * format, ...) {
	char outbuf[SESSION_LOG_LINE_SIZE];
	time_t current_time;
	va_list arglist;

	if (q_status.logging == Q_FALSE) {
		return;
	}

	time(&current_time);
	strftime(outbuf, sizeof(outbuf), "[%Y-%m-%d %H:%M:%S] ", localtime(&current_time));

	va_start(arglist, format);
	vsprintf((char *)(outbuf+strlen(outbuf)), format, arglist);
	va_end(arglist);

	fprintf(q_status.logging_file, outbuf);
	fflush(q_status.logging_file);
} /* ---------------------------------------------------------------------- */

/*
 * process_incoming_data - grab data from the child tty
 */
static void process_incoming_data() {
	int rc;				/* read() */
	int n;
	int unprocessed_n;
	char time_string[SHORT_TIME_SIZE];
	time_t current_time;
	int hours, minutes, seconds;
	double connect_time;
	int status;
	int i;
	char notify_message[DIALOG_MESSAGE_SIZE];

	Q_BOOL wait_on_script = Q_FALSE;

	/*
	 * For scripts: don't read any more data from the remote side if there
	 * is no more room in the print buffer side.
	 */
	if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
		if (q_running_script.print_buffer_full == Q_TRUE) {
			wait_on_script = Q_TRUE;
		}
	}
	
	if (((q_status.serial_open == Q_TRUE) || (q_status.online == Q_TRUE)) && FD_ISSET(q_child_tty_fd, &readfds)
		&& (wait_on_script == Q_FALSE)) {

		/*
		 * I read() from q_child_tty_fd into q_raw_buffer, then depending
		 * on my state I call another function to handle it.
		 */

		n = Q_BUFFER_SIZE - q_buffer_raw_n;

#ifdef DEBUG_IO
		fprintf(DEBUG_IO_HANDLE, "n = %d\n", n);
#endif

		if (n > 0) {

			/* Clear errno */
			errno = 0;

			rc = read(q_child_tty_fd, q_buffer_raw + q_buffer_raw_n, n);

#ifdef DEBUG_IO
			fprintf(DEBUG_IO_HANDLE, "rc = %d errno=%d\n", rc, errno);
#endif

			if (rc < 0) {
				if (errno == EIO) {
					/* This is EOF */
					rc = 0;
				} else {
					snprintf(notify_message, sizeof(notify_message), _("Call to read() failed: %s"), strerror(errno));
					notify_form(notify_message, 0);
					return;
				}
			}
			if (rc == 0) {
				/* EOF */
				if (q_status.serial_open) {
					/* Modem/serial */
					close_serial_port();
				} else {
					/* All other */
					q_status.online = Q_FALSE;
					/* Close pty */
					close(q_child_tty_fd);
					q_child_tty_fd = -1;
					Xfree(q_child_ttyname, __FILE__, __LINE__);
					wait4(q_child_pid, &status, WNOHANG, NULL);
					if (WIFEXITED(status)) {
						qlog(_("Connection exited with RC=%u\n"), WEXITSTATUS(status));
					} else if (WIFSIGNALED(status)) {
						qlog(_("Connection exited with signal=%u\n"), WTERMSIG(status));
					}
					if (q_current_dial_entry != NULL) {
						q_current_dial_entry->times_on++;
						time(&q_current_dial_entry->last_call);
					}
					q_child_pid = -1;
				}

				/* Compute time */
				/* time_string needs to be hours/minutes/seconds CONNECTED */
				time(&current_time);
				connect_time = difftime(current_time, q_status.connect_time);
				hours = connect_time / 3600;
				minutes = ((int)connect_time % 3600) / 60;
				seconds = (int)connect_time % 60;
				snprintf(time_string, sizeof(time_string), "%02u:%02u:%02u", hours, minutes, seconds);

				qlog(_("CONNECTION CLOSED. Total time online: %s\n"), time_string);
				/* Wipe out current dial entry */
				q_current_dial_entry = NULL;
				return;
			}

			time(&data_time);

			/* Mark/space parity */
			if ((q_status.serial_open == Q_TRUE) && (q_serial_port.parity == PARITY_MARK)) {
				/* Incoming data as MARK parity:  strip the 8th bit */
				for (i = 0; i < rc; i++) {
					q_buffer_raw[q_buffer_raw_n + i] &= 0x7F;
				}
			}

#ifdef LINE_NOISE
			for (i = 0; i < rc; i++) {
				int do_noise = random() % line_noise_per_bytes;
				if ((do_noise == 1) && (noise_stop == Q_FALSE)) {
					q_buffer_raw[i] = random() % 0xFF;
					noise_stop = Q_TRUE;
					break;
				}
			}
#endif
			
			/* Record # of new bytes in */
			q_buffer_raw_n += rc;

#ifdef DEBUG_IO
			fprintf(DEBUG_IO_HANDLE, "INPUT bytes: ");
			for (i=0; i<q_buffer_raw_n; i++) {
				fprintf(DEBUG_IO_HANDLE, "%02x ", q_buffer_raw[i] & 0xFF);
			}
			fprintf(DEBUG_IO_HANDLE, "\n");
#endif

		} /* if (n > 0) */
	} /* if (FD_ISSET(q_child_tty_fd, &readfds)) */

#ifdef DEBUG_IO
	fprintf(DEBUG_IO_HANDLE, "\n");
	fprintf(DEBUG_IO_HANDLE, "q_program_state: %d q_transfer_buffer_raw_n %d\n", q_program_state, q_transfer_buffer_raw_n);
	if (q_transfer_buffer_raw_n > 0) {
		fprintf(DEBUG_IO_HANDLE, "LEFTOVER OUTPUT\n");
	}
	fprintf(DEBUG_IO_HANDLE, "\n");
#endif

	unprocessed_n = q_buffer_raw_n;

	/* Modem dialer - allow everything to be sent first before looking for more data */
	if ((q_program_state == Q_STATE_DIALER) && (q_transfer_buffer_raw_n == 0)) {

		if (q_dial_state == Q_DIAL_CONNECTED) {
			/*
			 * UGLY HACK TIME.  Just so that I can display a message
			 * in the Redialer window, I duplicate the call to
			 * console_process_incoming_data() below.
			 */

			/* Let the console process the data */
			console_process_incoming_data(q_buffer_raw, q_buffer_raw_n, &unprocessed_n);

		} else {
			/* Normal case */
			dialer_process_data(q_buffer_raw, q_buffer_raw_n, &unprocessed_n, q_transfer_buffer_raw, &q_transfer_buffer_raw_n, sizeof(q_transfer_buffer_raw));
		}
	}

	/* File transfer */
	if (    (q_program_state == Q_STATE_UPLOAD) ||
		(q_program_state == Q_STATE_UPLOAD_BATCH) ||
		(q_program_state == Q_STATE_DOWNLOAD) ||
		(q_program_state == Q_STATE_SCRIPT_EXECUTE)
	) {


		/*
		 * File transfers and scripts: run protocol_process_data()
		 * until old_q_transfer_buffer_raw_n == q_transfer_buffer_raw_n .
		 *
		 * Every time we come through process_incoming_data() we call
		 * protocol_process_data() at least once.
		 */

		int old_q_transfer_buffer_raw_n = -1;
#ifdef DEBUG_IO
		fprintf(DEBUG_IO_HANDLE, "ENTER TRANSFER LOOP\n");
#endif

		while (old_q_transfer_buffer_raw_n != q_transfer_buffer_raw_n) {
			unprocessed_n = q_buffer_raw_n;
			old_q_transfer_buffer_raw_n = q_transfer_buffer_raw_n;

#ifdef DEBUG_IO
			fprintf(DEBUG_IO_HANDLE, "2 old_q_transfer_buffer_raw_n %d q_transfer_buffer_raw_n %d unprocessed_n %d\n", old_q_transfer_buffer_raw_n, q_transfer_buffer_raw_n, unprocessed_n);
#endif

			
			if (    (q_program_state == Q_STATE_UPLOAD) ||
				(q_program_state == Q_STATE_UPLOAD_BATCH) ||
				(q_program_state == Q_STATE_DOWNLOAD)) {

				/* File transfer protocol data handler */
				protocol_process_data(q_buffer_raw,
					q_buffer_raw_n, &unprocessed_n,
					q_transfer_buffer_raw, &q_transfer_buffer_raw_n,
					sizeof(q_transfer_buffer_raw));
			} else if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
				/* Script data handler */
				script_process_data(q_buffer_raw,
					q_buffer_raw_n, &unprocessed_n,
					q_transfer_buffer_raw, &q_transfer_buffer_raw_n,
					sizeof(q_transfer_buffer_raw));
				/* Reset the flags so the second call is a timeout type */
				q_running_script.stdout_readable = Q_FALSE;
				q_running_script.stdin_writeable = Q_FALSE;
			}

#ifdef DEBUG_IO
			fprintf(DEBUG_IO_HANDLE, "3 old_q_transfer_buffer_raw_n %d q_transfer_buffer_raw_n %d unprocessed_n %d\n", old_q_transfer_buffer_raw_n, q_transfer_buffer_raw_n, unprocessed_n);
#endif

			/* Hang onto whatever was unprocessed */
			if (unprocessed_n > 0) {
				memmove(q_buffer_raw, q_buffer_raw + q_buffer_raw_n - unprocessed_n, unprocessed_n);
			}
			q_buffer_raw_n = unprocessed_n;

#ifdef DEBUG_IO
			fprintf(DEBUG_IO_HANDLE, "4 old_q_transfer_buffer_raw_n %d q_transfer_buffer_raw_n %d unprocessed_n %d\n", old_q_transfer_buffer_raw_n, q_transfer_buffer_raw_n, unprocessed_n);
#endif

		}

#ifdef DEBUG_IO
		fprintf(DEBUG_IO_HANDLE, "EXIT TRANSFER LOOP\n");
#endif

	}

	/* Terminal mode */
	if (q_program_state == Q_STATE_CONSOLE) {
#ifdef DEBUG_IO
		fprintf(DEBUG_IO_HANDLE, "console_process_incoming_data: > q_buffer_raw_n %d unprocessed_n %d\n",
			q_buffer_raw_n, unprocessed_n);
#endif
		/* Let the console process the data */
		console_process_incoming_data(q_buffer_raw, q_buffer_raw_n, &unprocessed_n);

#ifdef DEBUG_IO
		fprintf(DEBUG_IO_HANDLE, "console_process_incoming_data: < q_buffer_raw_n %d unprocessed_n %d\n",
			q_buffer_raw_n, unprocessed_n);
#endif
	}

	assert(q_transfer_buffer_raw_n >= 0);
	assert(unprocessed_n >= 0);

	/* Hang onto whatever was unprocessed */
	if (unprocessed_n > 0) {
		memmove(q_buffer_raw, q_buffer_raw + q_buffer_raw_n - unprocessed_n, unprocessed_n);
	}
	q_buffer_raw_n = unprocessed_n;

#ifdef DEBUG_IO
	fprintf(DEBUG_IO_HANDLE, "serial_open = %s online = %s q_transfer_buffer_raw_n = %d\n",
		(q_status.serial_open == Q_TRUE ? "true" : "false"),
		(q_status.online == Q_TRUE ? "true" : "false"),
		q_transfer_buffer_raw_n);
#endif

	/* Write the data in the output buffer to q_child_tty_fd */
	if (((q_status.serial_open == Q_TRUE) || (q_status.online == Q_TRUE)) && (q_transfer_buffer_raw_n > 0)) {

#ifdef LINE_NOISE
		for (i = 0; i < q_transfer_buffer_raw_n; i++) {
			int do_noise = random() % line_noise_per_bytes;
			if ((do_noise == 1) && (noise_stop == Q_FALSE)) {
				q_transfer_buffer_raw[i] = random() % 0xFF;
				noise_stop = Q_TRUE;
				break;
			}
		}
#endif

		/* Write the data to q_child_tty_fd */
		rc = qodem_write(q_child_tty_fd, (char *)q_transfer_buffer_raw, q_transfer_buffer_raw_n, Q_TRUE);
		if (rc < 0) {

			switch (errno) {

			case EAGAIN:
				/* Outgoing buffer is full, wait for the next round */
				break;
			default:
				/* Uh-oh, error */
				snprintf(notify_message, sizeof(notify_message), _("Call to write() failed: %s"), strerror(errno));
				notify_form(notify_message, 0);
				return;
			}
		} else {
#ifdef DEBUG_IO
			fprintf(DEBUG_IO_HANDLE, "%d bytes written\n", rc);
#endif
			/* Hang onto the difference for the next round */
			assert(rc <= q_transfer_buffer_raw_n);

			if (rc < q_transfer_buffer_raw_n) {
				memmove(q_transfer_buffer_raw, q_transfer_buffer_raw + rc, q_transfer_buffer_raw_n - rc);
			}
			q_transfer_buffer_raw_n -= rc;
		}
	}

} /* ---------------------------------------------------------------------- */

/*
 * data_handler - block on data from either the control TTY or the data TTY
 */
static void data_handler() {
	int rc;				/* select() */
	int select_fd_max;		/* Maximum fd descriptor # */
	struct timeval listen_timeout;	/* select() timeout */
	time_t current_time;
	int default_timeout;		/* Microseconds to wait for data */
	char notify_message[DIALOG_MESSAGE_SIZE];
	char time_string[SHORT_TIME_SIZE];
	int hours, minutes, seconds;
	double connect_time;

#if 0

	/* Verify that q_child_tty_fd is still valid */
	struct stat fstats;
	if (q_child_tty_fd != -1) {
		if (fstat(q_child_tty_fd, &fstats) < 0) {
			/* "Oh shit" */
			assert (1 == 0);
		}
	}

#endif

	/*
	 * Default is to block 20 milliseconds (50Hz).
	 */
	default_timeout = 20000;

	/* Initialize select() structures */
	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	FD_ZERO(&exceptfds);

	/* Add stdin */
	select_fd_max = STDIN_FILENO;
	FD_SET(STDIN_FILENO, &readfds);

	/* Add the child tty */
	if (q_child_tty_fd != -1) {
		FD_SET(q_child_tty_fd, &readfds);

		/* Flag if we need to send data out to the child tty */
		if (q_transfer_buffer_raw_n > 0) {
			FD_SET(q_child_tty_fd, &writefds);
		}

		if (q_child_tty_fd > STDIN_FILENO) {
			select_fd_max = q_child_tty_fd;
		}
	}

	if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
		/* Add the script tty fd */
		if ((q_running_script.script_tty_fd != -1) && (q_running_script.paused == Q_FALSE)) {
			FD_SET(q_running_script.script_tty_fd, &readfds);
		
			/* Only check for writeability if we have something to send */
			if (q_running_script.print_buffer_empty == Q_FALSE) {
				FD_SET(q_running_script.script_tty_fd, &writefds);
			}

			if (q_running_script.script_tty_fd > select_fd_max) {
				select_fd_max = q_running_script.script_tty_fd;
			}
		}
	}
	
	/* select() needs 1 + MAX */
	select_fd_max++;

	/* Set the timeout */
	if (default_timeout >= 0) {
		/* Non-blocking select() */
		listen_timeout.tv_sec = default_timeout / 1000000;
		listen_timeout.tv_usec = default_timeout % 1000000;
		rc = select(select_fd_max, &readfds, &writefds, &exceptfds, &listen_timeout);
	} else {
		/* Blocking select() */
		rc = select(select_fd_max, &readfds, &writefds, &exceptfds, NULL);
	}

	/* fprintf(DEBUG_IO_HANDLE, "q_program_state = %d select() returned %d\n", q_program_state, rc); */
	switch (rc) {

	case -1:
		/* ERROR */
		switch (errno) {
		case EINTR:
			/* Interrupted system call, say from a SIGWINCH */
			break;
		default:
			snprintf(notify_message, sizeof(notify_message), _("Call to select() failed: %s"), strerror(errno));
			notify_form(notify_message, 0);
			exit(EXIT_ERROR_SELECT_FAILED);
		}
		break;
	case 0:
		/* Timeout, do nothing */

		/* Flush capture file if necessary */
		if (q_status.capture == Q_TRUE) {
			if (q_status.capture_flush_time < time(NULL)) {
				fflush(q_status.capture_file);
				q_status.capture_flush_time = time(NULL);
			}
		}

		/* Check for DCD drop */
		if ((q_status.online == Q_TRUE) && (q_status.serial_open == Q_TRUE)) {
			query_serial_port();
			if (q_serial_port.rs232.DCD == Q_FALSE) {
				qlog(_("OFFLINE: modem DCD line went down, lost carrier\n"));

				time(&current_time);
				connect_time = difftime(current_time, q_status.connect_time);
				hours = connect_time / 3600;
				minutes = ((int)connect_time % 3600) / 60;
				seconds = (int)connect_time % 60;
				snprintf(time_string, sizeof(time_string), "%02u:%02u:%02u", hours, minutes, seconds);

				qlog(_("CONNECTION CLOSED. Total time online: %s\n"), time_string);

				/* Modem/serial */
				close_serial_port();
			}
		}

		if ((q_child_tty_fd != -1) && (q_status.idle_timeout > 0)) {
			/* See if idle timeout was specified */
			time(&current_time);
			if ((difftime(current_time, data_time) > q_status.idle_timeout) && (difftime(current_time, q_data_sent_time) > q_status.idle_timeout)) {
				qlog(_("Connection IDLE timeout exceeded, closing...\n"));
				if (q_status.serial_open) {
					/* Modem/serial */
					close_serial_port();
				} else {
					/*
					 * All other cases.  We send the kill now,
					 * and the rest will be handled in
					 * process_incoming_data() .
					 */
					kill(q_child_pid, SIGHUP);
				}
			}
		}

		if ((q_child_tty_fd != -1) && (q_keepalive_timeout > 0) && (q_program_state != Q_STATE_DIALER)) {
			/* See if keepalive timeout was specified */
			time(&current_time);
			if ((difftime(current_time, data_time) > q_keepalive_timeout) && (difftime(current_time, q_data_sent_time) > q_keepalive_timeout)) {
				/* Send keepalive bytes */
				if (q_keepalive_bytes_n > 0) {
					qodem_write(q_child_tty_fd, q_keepalive_bytes, q_keepalive_bytes_n, Q_TRUE);
				}
			}
		}

		/*
		 * File transfer protocols and scripts need to run all the time.
		 */
		if ((q_program_state == Q_STATE_DOWNLOAD) ||
			(q_program_state == Q_STATE_UPLOAD) ||
			(q_program_state == Q_STATE_UPLOAD_BATCH) ||
			(q_program_state == Q_STATE_DIALER) ||
			(q_program_state == Q_STATE_SCRIPT_EXECUTE)
		) {
			/*
			 * For scripts: this is timeout, so don't try to move data
			 * to the pty/pipe.
			 */
			if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
				q_running_script.stdout_readable = Q_FALSE;
				q_running_script.stdin_writeable = Q_FALSE;
			}

			/* Process incoming data */
			process_incoming_data();
		}
		break;
	default:
		/*
		 fprintf(DEBUG_IO_HANDLE, "q_child_tty %s %s %s\n",
		 (FD_ISSET(q_child_tty_fd, &readfds) ? "READ" : ""),
		 (FD_ISSET(q_child_tty_fd, &writefds) ? "WRITE" : ""),
		 (FD_ISSET(q_child_tty_fd, &exceptfds) ? "EXCEPT" : ""));
		 */

		/* For scripts: see if stdout/stderr are readable and set flag */
		if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
			if (q_running_script.script_tty_fd != -1) {
				if (FD_ISSET(q_running_script.script_tty_fd, &readfds)) {
					q_running_script.stdout_readable = Q_TRUE;
				} else {
					q_running_script.stdout_readable = Q_FALSE;
				}
				if (FD_ISSET(q_running_script.script_tty_fd, &writefds)) {
					q_running_script.stdin_writeable = Q_TRUE;
				} else {
					q_running_script.stdin_writeable = Q_FALSE;
				}
			}
		}

		/* Data is present somewhere, go process it */
		if (	((q_child_tty_fd > 0) && (FD_ISSET(q_child_tty_fd, &readfds))) ||
			((q_child_tty_fd > 0) && (FD_ISSET(q_child_tty_fd, &writefds))) ||
			(q_program_state == Q_STATE_SCRIPT_EXECUTE)
		) {
			/* Process incoming data */
			process_incoming_data();
		}

		/* Note that keyboard_handler() will be called from
		 main() anyway, so just break here. */

		break;
	}

} /* ---------------------------------------------------------------------- */

/*
 * open_workingdir_file - open a file in the working dir.  *new_filename
 *                        will point to a new string if filename pointed
 *                        to a relative path, else *new_filename will
 *                        equal filename.
 */
FILE * open_workingdir_file(const char * filename, char ** new_filename) {
	*new_filename = NULL;

	if (strlen(filename) == 0) {
		return NULL;
	}

	if (filename[0] != '/') {
		/* Relative path, prefix working directory */
		*new_filename = (char *)Xmalloc(strlen(filename) +
			strlen(get_option(Q_OPTION_WORKING_DIR)) + 2, __FILE__, __LINE__);
		memset(*new_filename, 0, strlen(filename) + strlen(get_option(Q_OPTION_WORKING_DIR)) + 2);

		strncpy(*new_filename, get_option(Q_OPTION_WORKING_DIR), strlen(get_option(Q_OPTION_WORKING_DIR)));
		(*new_filename)[strlen(*new_filename)] = '/';
		strncpy(*new_filename + strlen(*new_filename), filename, strlen(filename));
	} else {
		/* Throw away the const */
		*new_filename = (char *)filename;
	}

	return fopen(*new_filename, "a");
} /* ---------------------------------------------------------------------- */

static char datadir_filename[FILENAME_SIZE];

/*
 * get_datadir_filename - open a file in the data dir.  *new_filename
 *                        will point to a new string if filename pointed
 *                        to a relative path, else *new_filename will
 *                        equal filename.
 */
char * get_datadir_filename(const char * filename) {
	assert(q_home_directory != NULL);

	sprintf(datadir_filename, "%s/%s", q_home_directory, filename);

	return datadir_filename;
} /* ---------------------------------------------------------------------- */

/*
 * get_workingdir_filename - open a file in the working directory.  *new_filename
 *                        will point to a new string if filename pointed
 *                        to a relative path, else *new_filename will
 *                        equal filename.
 */
char * get_workingdir_filename(const char * filename) {
	sprintf(datadir_filename, "%s/%s", get_option(Q_OPTION_WORKING_DIR), filename);

	return datadir_filename;
} /* ---------------------------------------------------------------------- */

/*
 * get_scriptdir_filename - open a file in the data dir.  *new_filename
 *                        will point to a new string if filename pointed
 *                        to a relative path, else *new_filename will
 *                        equal filename.
 */
char * get_scriptdir_filename(const char * filename) {
	assert(q_home_directory != NULL);

	sprintf(datadir_filename, "%s/scripts/%s", q_home_directory, filename);

	return datadir_filename;
} /* ---------------------------------------------------------------------- */

/*
 * open_datadir_file - open a file in the data dir.  *new_filename
 *                     will point to a new string if filename pointed
 *                     to a relative path, else *new_filename will
 *                     equal filename.
 */
FILE * open_datadir_file(const char * filename, char ** new_filename, const char * mode) {
	assert(q_home_directory != NULL);

	*new_filename = NULL;

	if (strlen(filename) == 0) {
		return NULL;
	}

	if (filename[0] != '/') {
		/* Relative path, prefix data directory */
		*new_filename = (char *)Xmalloc(strlen(filename) +
			strlen(q_home_directory) + 2, __FILE__, __LINE__);
		memset(*new_filename, 0, strlen(filename) + strlen(q_home_directory) + 2);

		strncpy(*new_filename, q_home_directory, strlen(q_home_directory));
		(*new_filename)[strlen(*new_filename)] = '/';
		strncpy(*new_filename + strlen(*new_filename), filename, strlen(filename));
	} else {
		/* Throw away the const */
		*new_filename = (char *)filename;
	}

	return fopen(*new_filename, mode);
} /* ---------------------------------------------------------------------- */

/*
 * main
 */
int main(int argc, char * const argv[]) {
	int option_index = 0;		/* getopt_long() */
	int rc;				/* getopt_long() */
	char * env_string;
	char * substituted_filename;
	wchar_t value_wchar[128];
	char * username;

#ifdef DEBUG_IO
	if (DEBUG_IO_HANDLE == NULL) {
		DEBUG_IO_HANDLE = fopen("debug_qodem.txt", "w");
	}
	fprintf(DEBUG_IO_HANDLE, "QODEM: START\n");
	fflush(DEBUG_IO_HANDLE);
#endif /* DEBUG_XMODEM */


	/* Internationalization */
#if defined(HAVE_LOCALE_H) || defined(HAVE_LC_MESSAGES)
	if (setlocale(LC_ALL, "") == NULL) {
		fprintf(stderr, "setlocale returned NULL: %s\n", strerror(errno));
		exit(EXIT_ERROR_SETLOCALE);
	}
#endif

#ifdef ENABLE_NLS
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	/*
	fprintf(stderr, "LANG: %s\n", getenv("LANG"));
	fprintf(stderr, "%s\n", bindtextdomain(PACKAGE, LOCALEDIR));
	fprintf(stderr, "%s\n", textdomain(PACKAGE));
	 */
#endif

	/* This is an UGLY HACK but it's somewhat important */
	rc = check_for_help(argc, argv);
	if (rc != 0) {
		exit(rc);
	}

	/* Initial call state */
	initial_call.method			= Q_DIAL_METHOD_SSH;
	initial_call.address			= NULL;
	username = getpwuid(geteuid())->pw_name;
	mbstowcs(value_wchar, username, strlen(username));
	initial_call.username = Xwcsdup(value_wchar, __FILE__, __LINE__);
	initial_call.password			= L"";
	initial_call.emulation			= Q_EMUL_LINUX;
	initial_call.codepage			= default_codepage(initial_call.emulation);
	initial_call.notes			= NULL;
	initial_call.script_filename		= NULL;
	initial_call.keybindings_filename	= NULL;
	initial_call.capture_filename		= NULL;
	initial_call.doorway			= Q_DOORWAY_CONFIG;
	initial_call.use_default_toggles	= Q_TRUE;

	/* Initial program state */
	q_program_state = Q_STATE_INITIALIZATION;
	/* Default to VT102 as the most common denominator */
	q_status.emulation		= Q_EMUL_VT102;
	q_status.codepage		= default_codepage(q_status.emulation);
	q_status.doorway_mode		= Q_DOORWAY_MODE_OFF;
	q_status.zmodem_autostart	= Q_TRUE;
	q_status.zmodem_escape_ctrl	= Q_FALSE;

	q_status.kermit_autostart		= Q_TRUE;
	q_status.kermit_robust_filename		= Q_FALSE;
	q_status.kermit_streaming		= Q_TRUE;
	q_status.kermit_uploads_force_binary	= Q_TRUE;
	q_status.kermit_downloads_convert_text	= Q_TRUE;

	q_status.serial_open		= Q_FALSE;
	q_status.online			= Q_FALSE;
	q_status.split_screen		= Q_FALSE;
	q_status.sound			= Q_FALSE;
	q_status.beeps			= Q_FALSE;
	q_status.ansi_music		= Q_FALSE;
	q_status.strip_8th_bit		= Q_FALSE;
	q_status.full_duplex		= Q_TRUE;
	q_status.line_feed_on_cr	= Q_FALSE;
	q_status.guard_hangup		= Q_TRUE;
	q_status.capture		= Q_FALSE;
	q_status.capture_file		= NULL;
	q_status.capture_type		= Q_CAPTURE_TYPE_NORMAL;
	q_status.capture_x		= 0;
	q_status.logging		= Q_FALSE;
	q_status.logging_file		= NULL;
	q_status.scrollback_enabled	= Q_TRUE;
	q_status.scrollback_lines	= 0;
	q_status.status_visible		= Q_TRUE;
	q_status.status_line_info	= Q_FALSE;
	q_status.hard_backspace		= Q_TRUE;
	/* Every console assumes line wrap, so turn it on by default */
	q_status.line_wrap		= Q_TRUE;
	/* BBS-like emulations usually assume 80 columns, so turn it on by default */
	q_status.assume_80_columns	= Q_TRUE;
	q_status.ansi_animate		= Q_FALSE;
	q_status.display_null		= Q_FALSE;
	q_status.reverse_video		= Q_FALSE;
	q_status.origin_mode		= Q_FALSE;
	q_status.insert_mode		= Q_FALSE;
	q_status.hold_screen_mode	= Q_FALSE;
	q_status.current_username	= NULL;
	q_status.current_password	= NULL;
	q_status.remote_address		= NULL;
	q_status.remote_port		= NULL;
	q_status.remote_phonebook_name	= NULL;
	q_status.idle_timeout		= 0;
	q_screensaver_timeout		= 0;
	q_keepalive_timeout		= 0;
	q_current_dial_entry		= NULL;

	/* Initialize the music "engine" :-) */
	music_init();

	/*
	 * Set q_home_directory.  load_options() will create the default key binding
	 * files and needs to use open_datadir_file().
	 */

	/* Sustitute for $HOME */
	env_string = getenv("HOME");

	/* Store for global use */
	q_home_directory = substitute_string("$HOME/.qodem", "$HOME", env_string);

	/* Load the options */
	load_options();

#ifndef QODEM_USE_SDL
#ifndef __linux
	/* Sound is not supported except on Linux console */
	q_status.sound = Q_FALSE;
#endif
#endif

	/* Setup MIXED mode doorway */
	setup_doorway_handling();

	/* Process options */
	for (;;) {
		rc = getopt_long(argc, argv, "h?", q_getopt_long_options, &option_index);
		if (rc == -1) {
			/* Fall out of the for loop, we're done calling getopt_long */
			break;
		}

		/* See which new option was specified */
		switch (rc) {
		case 0:
			process_command_line_option(q_getopt_long_options[option_index].name, optarg);
			break;
		case '?':
		case 'h':
			printf("%s", usage_string());
			q_program_state = Q_STATE_EXIT;
			break;
		default:
			break;
		}
	} /* for (;;) */

	/*
	 * Initialize the keyboard here.  It will newterm() each supported
	 * emulation, but restore things before it leaves.
	 */
	initialize_keyboard();

	/*
	 * Load the translation table
	 */
	load_translate_tables();

	/*
	 * Load the modem configuration
	 */
	load_modem_config();

	/* Initialize curses to setup stdscr */

	/*
	 * We reduce ESCDELAY on the assumption that local console
	 * is VERY fast.  However, if the user has already set ESCDELAY,
	 * we don't want to change it.
	 */
	if (getenv("ESCDELAY") == NULL) {
		putenv("ESCDELAY=20");
	}

	/* Xterm: send the private sequence to select metaSendsEscape */
	fprintf(stdout, "\033[?1036h");
	fflush(stdout);
	
	/* Initialize curses */
	screen_setup();

	if (strcasecmp(get_option(Q_OPTION_CAPTURE), "true") == 0) {
		start_capture(get_option(Q_OPTION_CAPTURE_FILE));
	}

	if (strcasecmp(get_option(Q_OPTION_LOG), "true") == 0) {
		start_logging(get_option(Q_OPTION_LOG_FILE));
	}

	/* Default scrolling region needs HEIGHT which is set by curses. */
	q_status.scroll_region_top	= 0;
	q_status.scroll_region_bottom	= HEIGHT - STATUS_HEIGHT - 1;

	/* Ignore SIGPIPE */
	signal(SIGPIPE, SIG_IGN);

	/*
	 * If anything else remains, turn it into a command line.
	 */
	if (optind < argc) {

		if ((strlen(argv[optind]) == 2) && (strcmp(argv[optind], "--") == 0)) {
			/*
			 * Special case:  strip out the "--" argument used to pass
			 * the remaining arguments to the connect program.
			 */
			optind++;
		}

		if (initial_call.address != NULL) {
			/* Error: --connect was specified along with a command line */
			screen_put_color_str_yx(0, 0,
				_("Error: The --connect argument cannot be used when a command"),  Q_COLOR_CONSOLE_TEXT);
			screen_put_color_str_yx(1, 0,
				_("line is also specified."),  Q_COLOR_CONSOLE_TEXT);
			/*
			 * Some X11 emulators (like GNOME-Terminal) will drop directly
			 * down to a command-line without leaving the usage info up.
			 * Force a keystroke so the user can actually see it.
			 */
			screen_put_color_str_yx(3, 0, _("Press any key to continue...\n"), Q_COLOR_CONSOLE_TEXT);
			screen_flush();
			discarding_getch();

			q_exitrc = EXIT_ERROR_COMMANDLINE;
			q_program_state = Q_STATE_EXIT;
			/*
			 * I'll let it finish constructing initial_call,
			 * but it'll never be used.
			 */
		}
		/* Set the dial method */
		initial_call.method = Q_DIAL_METHOD_COMMANDLINE;

		/* Build the command line */
		initial_call.address = Xmalloc(sizeof(char) * 1, __FILE__, __LINE__);
		initial_call.address[0] = '\0';
		for (;optind < argc; optind++) {
			/* Expand the buffer to include the next argument + one space +
			 null terminator. */
			initial_call.address = Xrealloc(initial_call.address, sizeof(char) * (strlen(initial_call.address) + strlen(argv[optind]) + 2), __FILE__, __LINE__);
			/* Zero out the new space */
			memset(initial_call.address + strlen(initial_call.address), 0, strlen(argv[optind]) + 1);
			/* Add the space before the next argument */
			initial_call.address[strlen(initial_call.address)] = ' ';
			/* Copy the next argument + the null terminator over */
			memcpy(initial_call.address + strlen(initial_call.address), argv[optind], strlen(argv[optind]) + 1);
		}
	} /* if (optind < argc) */

	if (q_program_state != Q_STATE_EXIT) {

#if 0

		/* This will show the curses keystroke for an input key */
		for (;;) {
			rc = getch();
			if (rc != -1) {
				printw("Key: %04o   %c\n", rc, (rc & 0xFF));
			}
			if (rc == 'q') {
				break;
			}
		}

#else

		/*
		 * Load the phonebook
		 */

		/* Sustitute for $HOME */
		env_string = getenv("HOME");

		/* Check for the default phonebook */
		substituted_filename = substitute_string("$HOME/.qodem/" DEFAULT_PHONEBOOK, "$HOME", env_string);
		if (access(substituted_filename, F_OK) != 0) {
			FILE * file;
			if ((file = fopen(substituted_filename, "w")) == NULL) {
				screen_put_color_printf_yx(0, 0, Q_COLOR_CONSOLE_TEXT, _("Error creating file \"%s\": %s\n"), substituted_filename, strerror(errno));
				screen_put_color_printf_yx(3, 0, Q_COLOR_CONSOLE_TEXT, _("Press any key to continue...\n"));
				screen_flush();
				discarding_getch();
			} else {
				fclose(file);
			}
		}
		q_phonebook.filename = substituted_filename;

		/* Now load it */
		load_phonebook(Q_FALSE);

		/*
		 * Explicitly call console_refresh() so that the scrollback will
		 * be set up.
		 */
		console_refresh(Q_FALSE);

		/* Reset all emulations */
		reset_emulation();

		if (initial_call.address != NULL) {
			q_current_dial_entry = &initial_call;
			q_status.remote_phonebook_name = Xmalloc(sizeof(wchar_t) * strlen(initial_call.address),
				__FILE__, __LINE__);
			mbstowcs(q_status.remote_phonebook_name, initial_call.address,
				strlen(initial_call.address) + 1);
			dial_out(&initial_call);
		} else if (strncmp(get_option(Q_OPTION_START_PHONEBOOK), "true", 4) == 0) {
			switch_state(Q_STATE_PHONEBOOK);
		} else {
			switch_state(Q_STATE_CONSOLE);
		}

		/* Enter main loop */
		for (;;) {
			/* Window size checks, refresh, etc. */
			refresh_handler();

			/* Grab data */
			data_handler();

			keyboard_handler();
			if (q_program_state == Q_STATE_EXIT) {
				break;
			}
		}

#endif

	} /* if (q_program_state != Q_STATE_EXIT) */

	/* Close any open files */
	stop_capture();
	stop_logging();
	if (q_status.serial_open == Q_TRUE) {
		close_serial_port();
	}

	/* Log our exit */
	qlog(_("Qodem exiting...\n"));

	/* Clear the screen */
	screen_clear();
	
	/* Shutdown curses */
	screen_teardown();

	/* Shutdown the music "engine" :-) */
	music_teardown();

#ifdef DEBUG_IO
	fclose(DEBUG_IO_HANDLE);
#endif

	/* Exit */
	exit(q_exitrc);
} /* ---------------------------------------------------------------------- */
